cpdef list gl_get_extensions()
cpdef int gl_has_extension(name)
cpdef gl_register_get_size(int constid, int size)
cpdef int gl_has_capability(int cap)
cpdef tuple gl_get_texture_formats()
cpdef int gl_has_texture_native_format(fmt)
cpdef int gl_has_texture_conversion(fmt)
cpdef int gl_has_texture_format(fmt)
cpdef tuple gl_get_version()
cpdef int gl_get_version_major()
cpdef int gl_get_version_minor()
