// Autogenerated file for Kivy C configuration
#define __PY3 1
#define __USE_RPI 0
#define __USE_EGL 0
#define __USE_OPENGL_ES2 0
#define __USE_OPENGL_MOCK 0
#define __USE_SDL2 1
#define __USE_PANGOFT2 0
#define __USE_IOS 0
#define __USE_ANDROID 0
#define __USE_MESAGL 0
#define __USE_X11 0
#define __USE_WAYLAND 0
#define __USE_GSTREAMER 0
#define __USE_AVFOUNDATION 0
#define __USE_OSX_FRAMEWORKS 0
#define __DEBUG_GL 0
#define __KIVY_SDL_GL_ALPHA_SIZE 8
