'''Kivy module for binary dependencies.

Binary dependencies such as gstreamer is installed as a
namespace module of kivy.deps. These modules are responsible
for making sure that the binaries are available to kivy.
'''
